from django.core.management.base import BaseCommand

from security.models import CeleryTaskLog, CeleryTaskLogState


class Command(BaseCommand):

    def handle(self, **options):
        stale_tasks = CeleryTaskLog.objects.filter_stale()
        if stale_tasks.exists():
            stale_tasks.update(
                state=CeleryTaskLogState.FAILED,
                error_message='Task execution was expired by command'
            )
            self.stderr.write(
                'Some tasks were too old and were set to error state. '
                'List of their IDs is:\n{}'.format(
                    ', '.join((str(v) for v in stale_tasks.values_list('pk', flat=True)))
                )
            )
        else:
            self.stdout.write('No stale tasks were found.')



import os
import base64
import logging
import pickle
import sys

from django.conf import settings
from django.core.management import call_command
from django.core.exceptions import ImproperlyConfigured
from django.db import transaction
from django.utils.timezone import now

try:
    from celery import Task
    from celery import shared_task
except ImportError:
    raise ImproperlyConfigured('Missing celery library, please install it')

from chamber.shortcuts import change_and_save
from chamber.utils.transaction import atomic_with_signals

from .models import CeleryTaskLog, CeleryTaskLogState


LOGGER = logging.getLogger(__name__)


class LoggedTask(Task):

    abstract = True
    logger_level = logging.WARNING

    def get_task(self, task_id):
        return CeleryTaskLog.objects.get(pk=task_id)

    def __call__(self, *args, **kwargs):
        # Every set attr is send here
        self.on_start(self.request.id, args, kwargs)
        super().__call__(*args, **kwargs)

    def _call_callback(self, event, *args, **kwargs):
        if hasattr(self, 'on_{}_callback'.format(event)):
            getattr(self, 'on_{}_callback'.format(event))(*args, **kwargs)

    def on_apply(self, task_id, args, kwargs):
        self._call_callback('apply', task_id, args, kwargs)

    def on_start(self, task_id, args, kwargs):
        self.get_task(task_id).change_and_save(state=CeleryTaskLogState.ACTIVE, start=now())
        self._call_callback('start', task_id, args, kwargs)

    def on_success(self, retval, task_id, args, kwargs):
        self.get_task(task_id).change_and_save(state=CeleryTaskLogState.SUCCEEDED, stop=now())
        self._call_callback('success', task_id, args, kwargs)

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        try:
            self.get_task(task_id).change_and_save(state=CeleryTaskLogState.FAILED, stop=now(), error_message=einfo)
        except CeleryTask.DoesNotExist:
            pass
        self._call_callback('failure', task_id, args, kwargs)

    def _create_task(self, options):
        task = CeleryTaskLog.objects.create(
            name=self.name, state=CeleryTaskLogState.WAITING,
            queue_name=options.get('queue', settings.CELERY_DEFAULT_QUEUE)
        )
        return str(task.pk)

    def _get_args(self, task_id, args):
        return (task_id,) + tuple(args or ())

    def apply_async_on_commit(self, args=None, kwargs=None, **options):
        task_id = self._create_task(options)
        args = self._get_args(task_id, args)
        self.on_apply(task_id, args, kwargs)
        if sys.argv[1:2] == ['test']:
            super().apply_async(args=args, kwargs=kwargs, task_id=task_id, **options)
        else:
            super_inst = super()
            transaction.on_commit(
                lambda: super_inst.apply_async(args=args, kwargs=kwargs, task_id=task_id, **options)
            )

    def apply_async(self, args=None, kwargs=None, **options):
        task_id = self._create_task(options)
        args = self._get_args(task_id, args)
        self.on_apply(task_id, args, kwargs)
        return super().apply_async(args=args, kwargs=kwargs, task_id=task_id, **options)

    def log_and_retry(self, attempt, exception_message=None, *args, **kwargs):
        LOGGER.log(self.logger_level, self.retry_error_message.format(
            attempt=attempt, exception_message=exception_message, **kwargs
        ))
        if attempt <= len(self.repeat_timeouts):
            self.apply_async_on_commit(
                args=args,
                kwargs={**kwargs, 'attempt': attempt+1},
                countdown=self.repeat_timeouts[attempt - 1] * 60,
                queue=self.queue
            )


def obj_to_string(obj):
    return base64.encodebytes(pickle.dumps(obj)).decode('utf8')


def string_to_obj(obj_string):
    return pickle.loads(base64.decodebytes(obj_string.encode('utf8')))


@shared_task(
    base=LoggedTask,
    bind=True,
    name='call_django_command',
)
@atomic_with_signals
def call_django_command(self, task_id, command_name, command_args=None):
    command_args = [] if command_args is None else command_args
    call_command(command_name, '--settings={}'.format(os.environ.get('DJANGO_SETTINGS_MODULE')), *command_args)




import shlex
import subprocess

from django.core.management.base import BaseCommand
from django.utils import autoreload


def restart_celery(celery_type, celery_settings, extra_arguments, pids):
    starter_celery_cmd = 'celery {} -l info -A {} {}'.format(celery_type, celery_settings, extra_arguments)
    kill_worker_cmd = 'pkill -9 -f "{}"'.format(starter_celery_cmd)
    subprocess.call(shlex.split(kill_worker_cmd))
    subprocess.call(shlex.split(starter_celery_cmd))


class Command(BaseCommand):

    def add_arguments(self, parser):

        parser.add_argument(
            'type', help='Celery type "beat" or "worker', choices={'beat', 'worker'}
        )
        parser.add_argument(
            '--celerysettings', dest='celery_settings', type=str,
            help='Tells Django to use celery settings', required=True
        )
        parser.add_argument(
            '--autoreload', action='store_true', dest='use_reloader',
            help='Tells Django to use the auto-reloader',
        )
        parser.add_argument(
            '--extra', dest='extra_args',
            help='Celery extra arguments"'
        )

    def handle(self, *args, **options):
        if options.get('use_reloader'):
            self.stdout.write('Starting celery with autoreload...')
            autoreload.main(restart_celery, args=(
                options.get('type'), options.get('celery_settings'), options.get('extra_args', ''), []
            ))
        else:
            self.stdout.write('Starting celery...')
            restart_celery(options.get('type'), options.get('celery_settings'), options.get('extra_args', ''), []),




from settings.celery import app as celery_app

from security.tasks import LoggedTask


@celery_app.task(
    base=LoggedTask,
    bind=True,
    name='sum_task')
def sum_task(self, task_id, a, b):
    return a + b


@celery_app.task(
    base=LoggedTask,
    bind=True,
    name='error_task')
def error_task(self, task_id):
    raise RuntimeError('error')



# Generated by Django 2.0 on 2019-08-15 13:08

from django.db import migrations, models
from django.db.models import F
import enumfields.fields
import security.models


def set_created_at_and_changed_at(apps, schema_editor):
    InputLoggedRequest = apps.get_model('security', 'InputLoggedRequest')
    OutputLoggedRequest = apps.get_model('security', 'OutputLoggedRequest')
    CommandLog = apps.get_model('security', 'CommandLog')

    InputLoggedRequest.objects.all().update(created_at=F('request_timestamp'), changed_at=F('request_timestamp'))
    OutputLoggedRequest.objects.all().update(created_at=F('request_timestamp'), changed_at=F('request_timestamp'))
    CommandLog.objects.all().update(created_at=F('start'), changed_at=F('start'))


class Migration(migrations.Migration):
    dependencies = [
        ('security', '0010_auto_20190710_1819'),
    ]

    operations = [
        migrations.CreateModel(
            name='CeleryTaskLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='created at')),
                ('changed_at', models.DateTimeField(auto_now=True, db_index=True, verbose_name='changed at')),
                ('start', models.DateTimeField(blank=True, null=True, verbose_name='start')),
                ('stop', models.DateTimeField(blank=True, null=True, verbose_name='stop')),
                ('name', models.CharField(max_length=250, verbose_name='task name')),
                ('state', enumfields.fields.NumEnumField(default=1, enum=security.models.CeleryTaskLogState,
                                                         verbose_name='state')),
                ('error_message', models.TextField(blank=True, null=True, verbose_name='error message')),
                ('queue_name', models.CharField(blank=True, max_length=250, null=True, verbose_name='queue name')),
            ],
            options={
                'verbose_name': 'celery task',
                'verbose_name_plural': 'celery tasks',
                'ordering': ('-created_at',),
            },
        ),
        migrations.AlterModelOptions(
            name='commandlog',
            options={'ordering': ('-created_at',), 'verbose_name': 'command log',
                     'verbose_name_plural': 'command logs'},
        ),
        migrations.AlterModelOptions(
            name='inputloggedrequest',
            options={'ordering': ('-created_at',), 'verbose_name': 'input logged request',
                     'verbose_name_plural': 'input logged requests'},
        ),
        migrations.AlterModelOptions(
            name='outputloggedrequest',
            options={'ordering': ('-created_at',), 'verbose_name': 'output logged request',
                     'verbose_name_plural': 'output logged requests'},
        ),
        migrations.AddField(
            model_name='commandlog',
            name='changed_at',
            field=models.DateTimeField(auto_now=True, db_index=True, null=True, blank=True,
                                       verbose_name='changed at'),
        ),
        migrations.AddField(
            model_name='commandlog',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, db_index=True, null=True, blank=True,
                                       verbose_name='created at'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='inputloggedrequest',
            name='changed_at',
            field=models.DateTimeField(auto_now=True, db_index=True, null=True, blank=True,
                                       verbose_name='changed at'),
        ),
        migrations.AddField(
            model_name='inputloggedrequest',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, db_index=True, null=True, blank=True,
                                       verbose_name='created at'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='outputloggedrequest',
            name='changed_at',
            field=models.DateTimeField(auto_now=True, db_index=True, null=True, blank=True,
                                       verbose_name='changed at'),
        ),
        migrations.AddField(
            model_name='outputloggedrequest',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, db_index=True, null=True, blank=True,
                                       verbose_name='created at'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='status',
            field=enumfields.fields.NumEnumField(default=0, enum=security.enums.LoggedRequestStatus,
                                                 verbose_name='status'),
        ),
        migrations.AlterField(
            model_name='inputloggedrequest',
            name='type',
            field=enumfields.fields.NumEnumField(default=1, enum=security.enums.InputLoggedRequestType,
                                                 verbose_name='type'),
        ),
        migrations.AlterField(
            model_name='outputloggedrequest',
            name='status',
            field=enumfields.fields.NumEnumField(default=0, enum=security.enums.LoggedRequestStatus,
                                                 verbose_name='status'),
        ),
        migrations.RunPython(set_created_at_and_changed_at),
    ]
